==> atmos-2.96.1.tar.gz: atmos-2.96.1/src/nr.cpp <==
/*
    $Id: nr.cpp 53 2009-11-01 18:03:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include "nr.h"

/**< nr.h is modified! erf*() are commented to avoid conflicts with math.h;
float->double; also all declarations of utilities are included there by:
#include "nrutil.h" in NR_CPP mode. The utilities are converted, to obey the
rules of C++
- declarations of external functions are removed from the function bodies
(instead, #include "nr.h" serves);
- arguments are supplied with their types in parentheses (nothing is thus in
between the arguments in parentheses and the function body {...},
- all the function names are prefixed with nr:: (i.e. belong to namespace "nr");
- all the calls to nrerror() are followed by instruction 'return'. */

void nr::four1( ublas::vector<double>& data  ) {
    int n,mmax,m,j,istep,i;
    double wtemp,wr,wpr,wpi,wi,theta;
    double tempr,tempi;

    n = data.size();
    j = 1;
    for( i = 1; i < n; i += 2 ) {
        if( j > i ) {
            std::swap( data(j-1), data(i-1) );
            std::swap( data(j),data(i) );
        }
        m = n/2;
        while( m >= 2 && j > m ) {
            j -= m;
            m /= 2;
        }
        j += m;
    }
    mmax = 2;
    while( n > mmax ) {
        istep = 2*mmax;
        theta = 2*M_PI/mmax;
        wtemp = sin( 0.5*theta );
        wpr = -2.0*wtemp*wtemp;
        wpi = sin( theta );
        wr = 1.0;
        wi = 0.0;
        for( m = 1; m < mmax; m += 2 ) {
            for( i = m; i <= n; i += istep ) {
                j = i + mmax;
                tempr = wr*data[j-1] - wi*data[j];
                tempi = wr*data[j] + wi*data[j-1];
                data[j-1] = data[i-1] - tempr;
                data[j] = data[i] - tempi;
                data[i-1] += tempr;
                data[i] += tempi;
            }
            wtemp = wr;
            wr = wr*wpr - wi*wpi + wr;
            wi = wi*wpr + wtemp*wpi + wi;
        }
        mmax = istep;
    }
}

void nr::hunt(double *xx, int n, double x, int *jlo)
{
    int jm,jhi,inc,ascnd;

    ascnd=(xx[n] > xx[1]);
    if (*jlo <= 0 || *jlo > n) {
        *jlo=0;
        jhi=n+1;
    } else {
        inc=1;
        if ((x >= xx[*jlo]) == ascnd) {
            if (*jlo == n) return;
            jhi=(*jlo)+1;
            while ((x >= xx[jhi]) == ascnd) {
                *jlo=jhi;
                inc += inc;
                jhi=(*jlo)+inc;
                if (jhi > n) {
                    jhi=n+1;
                    break;
                }
            }
        } else {
            if (*jlo == 1) {
                *jlo=0;
                return;
            }
            jhi=(*jlo);
            *jlo -= 1;
            while ((x < xx[*jlo]) == ascnd) {
                jhi=(*jlo);
                inc += inc;
                *jlo=jhi-inc;
                if (*jlo < 1) {
                    *jlo=0;
                    break;
                }
            }
        }
    }
    while (jhi-(*jlo) != 1) {
        jm=(jhi+(*jlo)) >> 1;
        if ((x > xx[jm]) == ascnd)
            *jlo=jm;
        else
            jhi=jm;
    }
}

#define FUNC(x) ((*func)(x))

double nr::midpnt(double (*func)(double), double a, double b, int n)
{
    double x,tnm,sum,del,ddel;
    static double s;
    static int it;
    int j;

    if (n == 1) {
        it=1;
        return (s=(b-a)*FUNC(0.5*(a+b)));
    } else {
        tnm=it;
        del=(b-a)/(3.0*tnm);
        ddel=del+del;
        x=a+0.5*del;
        sum=0.0;
        for (j=1;j<=it;j++) {
            sum += FUNC(x);
            x += ddel;
            sum += FUNC(x);
            x += del;
        }
        it *= 3;
        s=(s+(b-a)*sum/tnm)/3.0;
        return s;
    }
}

#undef FUNC

void nr::polint( double *xa, double *ya, int n, double x, double *y, double *dy) {
    int i,m,ns=1;
    double den,dif,dift,ho,hp,w;
    dif=fabs(x-xa[1]);
    ublas::vector<double> c(n+1);
    ublas::vector<double> d(n+1);
    for( i=1; i<=n; i++ ) {
        if ( (dift=fabs(x-xa[i])) < dif) {
            ns=i;
            dif=dift;
        }
        c(i)=ya[i];
        d(i)=ya[i];
    }
    *y=ya[ns--];
    for( m=1; m<n; m++ ) {
        for( i=1; i<=n-m; i++ ) {
            ho=xa[i]-x;
            hp=xa[i+m]-x;
            w=c(i+1)-d(i);
            den = ho - hp;
            assert( den != 0.0 );
            den = w/den;
            d(i) = hp*den;
            c(i) = ho*den;
        }
        *y += ( *dy = ( 2*ns < (n-m) ? c(ns+1) : d(ns--)) );
    }
}
/* This module contains the modified version of Numerical Recipe  qromo.c
   Modified by A. Tokovinin 27-Feb-97: if EPS>0 - absolute tolerance,
                                       if <0 - relative.
*/

#define JMAX 14
#define JMAXP JMAX+1
#define K 5
double nr::qromo(double (*func)(double), double a, double b,
                 double (*choose)(double (*)(double), double, double, int), double EPS ) {
    int j;
    double ss, dss, h[JMAXP+1], s[JMAXP+1];
    h[1] = 1.0;
    for( j = 1; j<= JMAX; j++ ) {
        s[j]=(*choose)( func, a, b, j);
        if( j >= K ) {
            polint( &h[j-K], &s[j-K], K, 0.0, &ss, &dss );
            if( EPS > 0 ) {
                if( std::abs(dss) < EPS) return ss;
            } else {
                if( std::abs(dss) < -EPS*std::abs(ss) ) return ss;
            }
        }
        s[j+1] = s[j];
        h[j+1] = h[j]/9.0;
    }
    return ss;
}

#undef JMAX
#undef JMAXP
#undef K

void nr::spline( const ublas::vector<double>& x, const ublas::vector<double>& y, double yp1, double ypn, ublas::vector<double>& y2 ) {
    assert(  x.size() ==  y.size() );
    double p, qn, sig, un;
    int n = x.size();
    ublas::vector<double> u( n );

    if( yp1 > 0.99e30 ) y2(0) = u(0) = 0.0;
    else {
        y2(0) = -0.5;
        u(0) = ( 3.0/(x(1)-x(0)) )*( (y(1)-y(0) )/( x(1)-x(0) ) - yp1 );
    }
    for( int i = 1; i < n-1; i++ ) {
        sig = ( x(i)-x(i-1) )/( x(i+1)-x(i-1) );
        p = sig*y2(i-1) + 2.0;
        y2(i) = ( sig - 1.0 )/p;
        u(i) = ( y(i+1) - y(i) )/( x(i+1)-x(i) ) - ( y(i) - y(i-1) )/( x(i)-x(i-1) );
        u(i) = ( 6.0*u(i)/( x(i+1)-x(i-1) ) - sig*u(i-1) )/p;
    }
    if( ypn > 0.99e30 ) qn = un = 0.0;
    else {
        qn = 0.5;
        un = ( 3.0/( x(n-1)-x(n-2) ) )*( ypn - ( y(n-1) - y(n-2) )/( x(n-1)-x(n-2) ));
    }
    y2(n-1) = ( un - qn*u(n-2) )/( qn*y2(n-2) + 1.0 );
    for( int k = n-2; k >= 0; k-- ) y2(k) = y2(k)*y2(k+1) + u(k);
}

void nr::splint( const ublas::vector<double>& xa, const ublas::vector<double>& ya, ublas::vector<double>& y2a, double x, double& y ) {
    int klo = 0;
    int khi = xa.size()-1;
    int k = 0;
    double h, b, a;
    while( khi - klo > 1 ) {
        k = ( khi + klo )/2;
        if( xa(k) > x ) khi = k;
        else klo = k;
    }
    h = xa(khi) - xa(klo);
    assert( h != 0 );
    a = ( xa(khi) - x )/h;
    b = ( x - xa(klo) )/h;
    y = a*ya(klo) + b*ya(khi) + (( a*a*a - a )*y2a(klo) + ( b*b*b - b )*y2a(khi) )*h*h/6.0;
}

==> atmos-2.96.7.tar.gz: atmos-2.96.7/src/nr.cpp <==
/*
    $Id: nr.cpp 101 2010-03-02 08:36:15Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include "nr.h"

/**< nr.h is modified! erf*() are commented to avoid conflicts with math.h;
float->double; also all declarations of utilities are included there by:
#include "nrutil.h" in NR_CPP mode. The utilities are converted, to obey the
rules of C++
- declarations of external functions are removed from the function bodies
(instead, #include "nr.h" serves);
- arguments are supplied with their types in parentheses (nothing is thus in
between the arguments in parentheses and the function body {...},
- all the function names are prefixed with nr:: (i.e. belong to namespace "nr");
- all the calls to nrerror() are followed by instruction 'return'. */

void nr::four1( ublas::vector<double>& data  ) {
    int n,mmax,m,j,istep,i;
    double wtemp,wr,wpr,wpi,wi,theta;
    double tempr,tempi;

    n = data.size();
    j = 1;
    for( i = 1; i < n; i += 2 ) {
        if( j > i ) {
            std::swap( data(j-1), data(i-1) );
            std::swap( data(j),data(i) );
        }
        m = n/2;
        while( m >= 2 && j > m ) {
            j -= m;
            m /= 2;
        }
        j += m;
    }
    mmax = 2;
    while( n > mmax ) {
        istep = 2*mmax;
        theta = 2*M_PI/mmax;
        wtemp = sin( 0.5*theta );
        wpr = -2.0*wtemp*wtemp;
        wpi = sin( theta );
        wr = 1.0;
        wi = 0.0;
        for( m = 1; m < mmax; m += 2 ) {
            for( i = m; i <= n; i += istep ) {
                j = i + mmax;
                tempr = wr*data[j-1] - wi*data[j];
                tempi = wr*data[j] + wi*data[j-1];
                data[j-1] = data[i-1] - tempr;
                data[j] = data[i] - tempi;
                data[i-1] += tempr;
                data[i] += tempi;
            }
            wtemp = wr;
            wr = wr*wpr - wi*wpi + wr;
            wi = wi*wpr + wtemp*wpi + wi;
        }
        mmax = istep;
    }
}

void nr::spline( const ublas::vector<double>& x, const ublas::vector<double>& y, double yp1, double ypn, ublas::vector<double>& y2 ) {
    assert(  x.size() ==  y.size() );
    double p, qn, sig, un;
    int n = x.size();
    ublas::vector<double> u( n );
    if( yp1 > 0.99e30 ) y2(0) = u(0) = 0.0;
    else {
        y2(0) = -0.5;
        u(0) = ( 3.0/(x(1)-x(0)) )*( (y(1)-y(0) )/( x(1)-x(0) ) - yp1 );
    }
    for( int i = 1; i < n-1; i++ ) {
        sig = ( x(i)-x(i-1) )/( x(i+1)-x(i-1) );
        p = sig*y2(i-1) + 2.0;
        y2(i) = ( sig - 1.0 )/p;
        u(i) = ( y(i+1) - y(i) )/( x(i+1)-x(i) ) - ( y(i) - y(i-1) )/( x(i)-x(i-1) );
        u(i) = ( 6.0*u(i)/( x(i+1)-x(i-1) ) - sig*u(i-1) )/p;
    }
    if( ypn > 0.99e30 ) qn = un = 0.0;
    else {
        qn = 0.5;
        un = ( 3.0/( x(n-1)-x(n-2) ) )*( ypn - ( y(n-1) - y(n-2) )/( x(n-1) - x(n-2) ));
    }
    y2(n-1) = ( un - qn*u(n-2) )/( qn*y2(n-2) + 1.0 );
    for( int k = n-2; k >= 0; k-- ) y2(k) = y2(k)*y2(k+1) + u(k);
}

void nr::splint( const ublas::vector<double>& xa, const ublas::vector<double>& ya, ublas::vector<double>& y2a, double x, double& y ) {
    int klo = 0;
    int khi = xa.size()-1;
    int k = 0;
    while( khi - klo > 1 ) {
        k = ( khi + klo )/2;
        if( xa(k) > x ) khi = k;
        else klo = k;
    }
    double h = xa(khi) - xa(klo);
    assert( h != 0 );
    double a = ( xa(khi) - x )/h;
    double b = ( x - xa(klo) )/h;
    y = a*ya(klo) + b*ya(khi) + (( a*a*a - a )*y2a(klo) + ( b*b*b - b )*y2a(khi) )*h*h/6.0;
}
